__all__ = ["EnrichImage", "ParentAsLabel"]

from .basic import Enrich
from tai_chi_tuna.front.typer import STR, LIST
from pathlib import Path
from PIL import Image
import json

class EnrichImage(Enrich):
    """
    Create Image column from image path column
    """
    prefer = "QuantifyImage"
    typing = Image
    lazy = True
    stateful_dict = dict(
        convert="str",
        size="list",
    )
    def __init__(
        self, convert: STR("RGB") = "RGB",
        size: LIST(options=[28, 128, 224, 256, 512], default=224) = 224,
    ):
        self.convert = convert
        self.size = size

    def __repr__(self):
        return f"[Image:{self.size}]"

    def __call__(self, x):
        img = Image.open(x).convert(self.convert)
        img = img.resize((self.size, self.size))
        return img


class ParentAsLabel(Enrich):
    typing = str
    prefer = "QuantifyCategory"
    def __call__(self, path: Path,) -> str:
        """
        Use parent folder name as label
        """
        return Path(path).parent.name


__all__ = ["CategoryTop", "MultiCategoryTop"]

from .metric import Accuracy, BiAccuracy, F1Score
from .loss import BCEWithLogitsLossCasted
from .basic import ExitModel, nn, HIDDEN_SIZE_OPTIONS


class CategoryTop(ExitModel):
    prefer = "CrossEntropyLoss"
    input_dim = 2

    def __init__(self, in_features, out_features, hidden_size):
        super().__init__()
        self.hidden_size = hidden_size
        self.create_top(in_features, out_features, hidden_size)
        self.activation = nn.Softmax(dim=-1)
        self.crit = nn.CrossEntropyLoss()
        self.metric_funcs.update(
            {
                "acc": Accuracy(),
            })

    def forward(self, x):
        return self.top(x)

    @classmethod
    def from_quantify(
        cls,
        quantify,
        entry_part,
        hidden_size: HIDDEN_SIZE_OPTIONS = 0,
    ):
        """
        Build the exit model parts
            with the quantify and entry parts
        """
        out_features = len(quantify.category)
        in_features = entry_part.out_features
        return cls(
            in_features=in_features,
            out_features=out_features,
            hidden_size=hidden_size,
        )


class MultiCategoryTop(ExitModel):
    prefer = "BCEWithLogitsLossCasted"
    input_dim = 2

    def __init__(
        self,
        in_features: int,
        out_features: int,
        hidden_size: int = 0
    ):
        super().__init__()
        self.create_top(in_features, out_features, hidden_size)
        self.activation = nn.Sigmoid()
        self.crit = BCEWithLogitsLossCasted()
        self.metric_funcs.update(
            {"acc": BiAccuracy(), "F1": F1Score()})

    def forward(self, x):
        return self.top(x)

    @classmethod
    def from_quantify(
            cls,
            quantify,
            entry_part,
            hidden_size: HIDDEN_SIZE_OPTIONS = 0,):
        """
        Build the exit model parts
            with the quantify and entry parts
        """
        out_features = len(quantify.category)
        in_features = entry_part.out_features
        return cls(
            in_features=in_features,
            out_features=out_features,
            hidden_size=hidden_size,
        )



from torch import nn
__all__ = ["BCEWithLogitsLossCasted", ]


class BCEWithLogitsLossCasted(nn.Module):
    def __init__(self):
        super().__init__()
        self.bce_ = nn.BCEWithLogitsLoss()

    def forward(self, y_, y):
        return self.bce_(y_, y.float())



__all__ = ["ImageConvEncoder"]

from tai_chi_tuna.front.typer import LIST
from tai_chi_tuna.flow.to_quantify import BATCH_SIZE

from .basic import EntryModel, Empty

from torchvision.models import (
    resnet18, resnet34, resnet50, resnet101, resnet152,
    resnext101_32x8d, resnext50_32x4d)

RESNET_OPTIONS = {"resnet18": resnet18,
                  "resnet34": resnet34,
                  "resnet50": resnet50,
                  "resnet101": resnet101,
                  "resnet152": resnet152,
                  "resnext101_32x8d": resnext101_32x8d,
                  "resnext50_32x4d": resnext50_32x4d}

class ImageConvEncoder(EntryModel):
    def __init__(self, model):
        super().__init__()
        self.name = "cnn"
        self.output_shape = (BATCH_SIZE, model.fc.in_features)
        self.out_features = model.fc.in_features
        model.fc = Empty()
        self.model = model

    def forward(self, data):
        return self.model(data)

    def __repr__(self):
        return f"""ComputerVisionEncoder: {self.name}
        Outputs shape:{self.output_shape}"""

    @classmethod
    def from_quantify(
        cls,
        quantify,
        name: LIST(options=list(
            RESNET_OPTIONS.keys()), default="resnet18"),
    ):
        # do not load pretrained for inference
        if quantify.is_inference:
            model = RESNET_OPTIONS[name](pretrained=False)
        # load pretrained for transfer learning
        else:
            model = RESNET_OPTIONS[name](pretrained=True, progress=True,)
        obj = cls(model)
        obj.name = name
        return obj


__all__ = ["EnrichImage", "ParentAsLabel", "ENRICHMENTS"]


from .basic import EnrichCleanTyping
from .file import (
    EnrichImage, ParentAsLabel
    )

ENRICHMENTS = dict(
    EnrichImage=EnrichImage,
    ParentAsLabel=ParentAsLabel,
    CleanData=EnrichCleanTyping
)


__all__ = ['EntryModel', 'Empty', 'ImageConvEncoder',
           'CategoryEncoder', 'MultiCategoryEncoder', 'TransformerEncoder',
           'ALL_ENTRY', 'QUANTIFY_2_ENTRY_MAP'
           ]

from .basic import (
    EntryModel, Empty,
)
from .image import (
    ImageConvEncoder,
)
from .category import (
    CategoryEncoder, MultiCategoryEncoder
)
from .text import (
    TransformerEncoder
)

from tai_chi_engine.quantify import (
    QuantifyImage, QuantifyText, 
    QuantifyCategory, QuantifyMultiCategory, QuantifyNum
)

ALL_ENTRY = dict(
    ImageConvEncoder=ImageConvEncoder,
    CategoryEncoder=CategoryEncoder,
    MultiCategoryEncoder=MultiCategoryEncoder,
    TransformerEncoder=TransformerEncoder,
    Empty=Empty,
)

QUANTIFY_2_ENTRY_MAP = dict({
    QuantifyImage:[
        ImageConvEncoder,
    ],
    QuantifyCategory:[
        CategoryEncoder,
    ],
    QuantifyMultiCategory:[
        MultiCategoryEncoder,
    ],
    QuantifyText:[
        TransformerEncoder,
    ],
    QuantifyNum:[
        Empty,
    ],
})


__all__ = ["ExitModel", "RegressionTop", "HIDDEN_SIZE_OPTIONS"]

from torch import nn
import numpy as np
from .metric import MeanAbsoluteError
from tai_chi_tuna.front.typer import LIST


class ExitModel(nn.Module):
    """
    # Base Model of exit part

    ## You **have to** set the following for the inherited class
    - self.crit: the criterion, a callble to calc ```self.crit(y_, y)```
    - self.from_quantify: a class method to build the model from entry_dict and quantify class

    ## You **can** set/ overwrite the following for the inherited class
    - self.metric_funcs: a dict of metric functions, key is the name of the metric
    - self.activation: the activation function, a callble to calc ```self.activation(y_)```
    - self.loss_step: a callble to calc ```self.loss_step(x, y)```
    - self.eval_forward: a callble to calc ```self.eval_forward(x)```
    """
    metric_funcs = dict()

    def loss_step(self, x, y):
        y_ = self(x)
        loss = self.crit(y_, y)
        metrics = dict()
        if hasattr(self, "activation"):
            y_ = self.activation(y_)
        # calculate metrics key by key
        for k, func in self.metric_funcs.items():
            metric = func(y_, y)
            if hasattr(metric, "keys"):
                metrics.update(metric)
            else:
                metrics[k] = metric
        return dict(loss=loss, y_=y_, **metrics)

    def create_top(
            self, in_features: int,
            out_features: int,
            hidden_size: int = 0) -> nn.Module:
        """
        Create the top part of the model
        if hidden_size is 0, then no hidden layer (only a linear layer)
        """
        if hidden_size > 0:
            self.top = nn.Sequential(
                nn.Linear(in_features=in_features, out_features=hidden_size),
                nn.BatchNorm1d(hidden_size),
                nn.ReLU(),
                nn.Linear(in_features=hidden_size, out_features=out_features),
            )
        else:
            self.top = nn.Linear(
                in_features=in_features, out_features=out_features)
        return self.top

    def eval_forward(self, x) -> np.ndarray:
        """
        Forward pass for 
        """
        y_ = self(x)
        if hasattr(self, "activation"):
            y_ = self.activation(y_)
        return y_

    @classmethod
    def from_quantify(cls, ):
        raise ImportError(
            f"Please define class function 'from_quantify' for {cls.__name__}"
        )


HIDDEN_SIZE_OPTIONS = LIST(default=0, options=[0, 64, 128, 256, 512, 1024, 2048])


class RegressionTop(ExitModel):
    prefer = "MSELoss"
    input_dim = 2

    def __init__(self, in_features, out_features, hidden_size=0):
        super().__init__()
        self.create_top(in_features, out_features, hidden_size)
        self.crit = nn.MSELoss()
        self.metric_funcs.update(
            {
                "mae": MeanAbsoluteError(),
            })

    def forward(self, x):
        return self.top(x)

    @classmethod
    def from_quantify(
        cls,
        quantify,
        entry_part,
        hidden_size: HIDDEN_SIZE_OPTIONS = 0):
        out_features = 1
        in_features = entry_part.out_features
        return cls(
            in_features=in_features,
            out_features=out_features,
            hidden_size=hidden_size,
        )



from tai_chi_tuna.config import PhaseConfig
from tai_chi_tuna.flow.to_quantify import load_qdict
from tai_chi_tuna.flow.to_model import assemble_model

from .enrich import ENRICHMENTS
from .quantify import QUANTIFY
from .model import (ALL_EXIT, QUANTIFY_2_EXIT_MAP,
                    ALL_ENTRY, QUANTIFY_2_ENTRY_MAP)
import torch
import logging
from pathlib import Path
from typing import Dict, Any


class TaiChiTrained:
    """
    Trained project
    """
    def __init__(self, project: Path, device:str = "cpu"):
        """
        Load a trained project from a project directory
        """
        self.project = Path(project)
        if self.project.exists() == False:
            raise FileNotFoundError(
                f"Project {self.project} does not exist")
        self.phase = PhaseConfig.load(project)
        self.load_things()
        self.device = torch.device(device)

    def __repr__(self):
        return f"[☯️ Project: {self.project}]\n" +\
            "\tmodel:\tself.final_model\n" +\
            "\tquantify:\tself.qdict\n" +\
            f"\tx_columns:\t{self.x_columns}\n" +\
            f"\ty_columns:\t{self.y_columns}\n"

    @property
    def best_checkpoint(self,):
        checkpoints = list((self.project/"checkpoints").glob("*.ckpt"))
        if len(checkpoints) == 0:
            raise FileNotFoundError(
                f"No checkpoints found in {self.project}")
        return checkpoints[-1]

    def to_tensor(self, data: Dict[str, Any]):
        """
        Convert the data to tensor
        """
        tensor_data = dict()
        for k in self.x_columns:
            value = data[k]
            tensor_data[k] = self.qdict[k](list([value,]))
            tensor_data[k].to(self.device)
        return tensor_data

    def predict(self, data):
        tensor_data = self.to_tensor(data)
        with torch.no_grad():
            pred = self.final_model.eval_forward(tensor_data)
        return self.y_quantify.backward(pred[0])

    def load_things(self):
        module_zoo = {"all_entry": ALL_ENTRY, "all_exit": ALL_EXIT}
        self.qdict = load_qdict(project=self.project,
                                phase=self.phase, quantify_map=QUANTIFY)

        self.final_model = assemble_model(
            phase=self.phase, qdict=self.qdict, modules=module_zoo)
            
        logging.info(f"Loaded model from {self.best_checkpoint}")
        state_dict = torch.load(str(self.best_checkpoint), map_location="cpu")
        if 'state_dict' in state_dict:
            state_dict = state_dict['state_dict']

        self.final_model.load_state_dict(state_dict)
        self.final_model.eval()

        self.x_columns = list(
            quantify['src']
            for quantify in self.phase['quantify'] if quantify['x'])

        self.y_columns = list(
            quantify['src']
            for quantify in self.phase['quantify'] if quantify['x']==False)

        self.y_quantify = self.qdict[self.y_columns[0]]



__all__ = ['EntryModel', 'Empty']

from torch import nn
from tai_chi_engine.stateful import Stateful


class EntryModel(nn.Module):
    is_entry = True
    
    @classmethod
    def from_quantify(cls, quantify):
        raise ImportError(
            f"Please define class function 'from_quantify' for {cls.__name__}"
        )
    
class Empty(EntryModel):
    def __init__(self):
        super().__init__()
        self.out_features=1
    
    def forward(self, x):
        return x
    
    @classmethod
    def from_quantify(cls,
        quantify):
        return cls()




__all__ = ['TransformerEncoder']

from tai_chi_tuna.front.typer import BOOL, STR
from .basic import EntryModel

class TransformerEncoder(EntryModel):
    """
    A model part to encode sequnce data in to vectors
    """

    def __init__(self, model, encoder_mode: BOOL(default=True) = True,):
        super().__init__()
        self.model = model
        self.encoder_mode = encoder_mode

    def reduce_hidden_state(self, vec, attention_mask):
        """
        Mean pooling over the hidden state according to attention_mask
        """
        vec = vec*(attention_mask[...,None])
        return vec.sum(dim=1)/(attention_mask.sum(-1)[...,None])

    def forward(self, kwargs):
        outputs = self.model(**kwargs)
        if self.encoder_mode:
            # output vector
            return self.reduce_hidden_state(
                outputs.last_hidden_state, kwargs["attention_mask"])
        return outputs

    @classmethod
    def from_quantify(
        cls,
        quantify,
        name: STR(default="bert-base-uncased") = 'bert-base-uncased',
        encoder_mode: BOOL(default=True) = True,
    ):
        # load transformer for inference
        if quantify.is_inference:
            from transformers import AutoConfig, AutoModel
            config = AutoConfig.from_pretrained(name)
            model = AutoModel.from_config(config)
        # load transformer for training
        else:
            from transformers import AutoModel
            model = AutoModel.from_pretrained(name)
        # substatiate the transformer model
        obj = cls(model)
        obj.name = name
        obj.encoder_mode = encoder_mode
        if encoder_mode:
            obj.out_features= model.config.hidden_size
        return obj


from tai_chi_tuna.utils import clean_name
from pathlib import Path
import pandas as pd

from tai_chi_tuna.config import PhaseConfig
from tai_chi_tuna.front.html import DOM, list_group_kv
from pathlib import Path

def df_creator_image_folder(path: Path) -> pd.DataFrame:
    """
    Create a dataframe ,
    Which list all the image path under a system folder
    """
    path = Path(path)
    files = []
    formats = ["jpg", "jpeg", "png"]
    for fmt in formats:
        files.extend(path.rglob(f"*.{fmt.lower()}"))
        files.extend(path.rglob(f"*.{fmt.upper()}"))
    return pd.DataFrame({"path": files}).sample(frac=1.).reset_index(drop=True)

def join_col_list(Columns):
    return ",".join(map(lambda x:f"'{x['src']}'", Columns))

class Narrator:
    """
    Under jupyter notebook: 
    ```python
    narrator = Narrator(phase:PhaseConfig)
    narrator()
    ```
    This will narrate how to train the model, in full detail
    """
    
    x_color = "#3399FF"
    y_color = "#33CC55"
    next_color = "#11CCDD"

    def __init__(self, phase: PhaseConfig):
        self.phase = phase

    def __repr__(self):
        return str(self)

    def __str__(self) -> str:
        return f"<Narrator>, run self() to print out in juypter"

    def __call__(self,):
        self.narrate()()

    def title(self, text) -> DOM:
        return DOM(text, "h3",
                   {"style": "color:#FFAA33;background-color:#FFFFFF;padding:3px"})

    def empty_step(self,) -> DOM:
        return DOM("", "ul", dict())

    def line(self, inner) -> DOM:
        return DOM(inner, "li", dict(style="padding:2px;background-color:#FFFFDD"))
    
    def ok(self, text) -> DOM:
        return DOM(text, "h5",
                   dict(style=f"background-color:{self.next_color};color:#FFF;padding:3px"))

    def narrate(self) -> DOM:
        """
        Chaining up 4 steps
        """
        doc = DOM("", "div")
        step1 = self.step_enrich()
        doc.append(step1)

        step2 = self.step_quantify()
        doc.append(step2)

        step3 = self.step_model()
        doc.append(step3)
        
        step4 = self.step_train()
        doc.append(step4)

        return doc

    def step_enrich(self) -> DOM:
        config = self.phase.config
        step = self.empty_step()
        step.append(self.title("Step 1: Enrich"))
        if "enrich" in config:
            enrichs = config['enrich']
            for enrich in enrichs:
                src = enrich.get('src')
                enrich_type = enrich.get('enrich')
                step.append(self.line(f"select column '{src}' for 'src'"))
                step.append(
                    self.line(
                        f"select '{enrich_type}' for 'enrich'"))
                step.append(
                    self.line(
                        f"The kwargs of '{enrich_type}' should be set to:"))
                step.append(self.line(list_group_kv(enrich.get("kwargs"))))
            step.append(
                self.ok("Congrats! now Click ✔️Next above"))
        else:
            step.append(
                self.ok("No enrich set, click ✔️Next above"))
        return step

    def step_quantify(self) -> DOM:
        config = self.phase.config
        step = self.empty_step()
        step.append(self.title("Step 2: Quantify"))
        if "quantify" in config:
            quantifies = config['quantify']
            Xs = list(filter(lambda x: x['x'], quantifies))
            Ys = list(filter(lambda x: x['x'] == False, quantifies))

            step.append(f"""
                In this model we are using 
                <strong style='color:{self.x_color}'>
                column(s):{join_col_list(Xs)}</strong> 
                to guess <strong style='color:{self.y_color}'>
                column:{join_col_list(Ys)}</strong>
                """)

            for is_x, Columns in zip([True, False], [Xs, Ys]):
                if is_x:
                    step.append(
                            f"""
                            <h4 style='color:{self.x_color}'>
                            There are {len(Xs)} X column(s):</h4>"""
                        )
                else:
                    step.append(
                            f"""
                            <h4 style='color:{self.y_color}'>
                            There is our {len(Ys)} Y column:</h4>"""
                        )
                for col in Columns:
                    step.append(
                        self.line(f"""
                        <h5>Quantify config for '{col['src']}' column</h5>""")
                    )
                    step.append(
                        self.line(f"""
                        Select '{col['src']}' as 'src'"""))
                    step.append(
                        self.line(f"""
                        Choose 'As {'X' if is_x else 'Y'}', 
                        So column {col['src']} is used as 
                        {'input' if is_x else 'output'} data"""))
                    step.append(
                        self.line(f"Click Button: 'Run Interact'"))
                    step.append(
                        self.line(f"""
                        Choose '{col['quantify']}' as the method to 
                        transform '{col['src']}' into matrix/tensor"""))
                    if 'kwargs' in col:
                        step.append(
                            self.line(
                                f"Click Button: 'Run Interact' to set kwargs"))
                        step.append(
                            self.line(
                                f"The kwargs of '{col['quantify']}' should be set to:"))
                        step.append(
                            self.line(list_group_kv(col["kwargs"])))
                        step.append(
                            self.line(f'Click Button:+Create'))
                    else:
                        step.append(
                            self.line(f"""
                            Click Button: 'Run Interact' then click '+Create'"""))
            step.append(
                self.ok("Congrats! now Click ✔️Next above")
            )
        return step

    def step_model(self) -> DOM:
        config = self.phase.config
        step = self.empty_step()
        step.append(self.title("Step 3: Model"))
        step.append("<h5>Batching up Data</h5>")

        batch_level = config['batch_level']
        valid_ratio = batch_level['valid_ratio']
        batch_size = batch_level['batch_size']
        shuffle = 'checked' if batch_level['shuffle'] else 'unchecked'
        num_workers = batch_level['num_workers']
        step.append(
            self.line(
                f"""We slide valid_ratio to '{valid_ratio}'
                    as around '{int(valid_ratio*100)}%' of total data
                    will be used as validation set""")
        )
        step.append(
            self.line(f"""
            We set batch_size to '{batch_size}',
            if you run into memory error or GPU OOM,
            keep halfing this number, until it works""")
        )
        step.append(
            self.line(f"We keep the 'shuffle' checkbox to '{shuffle}'")
        )
        step.append(
            self.line(
                f"""We set the 'num_workers' to '{num_workers}',
                so {1 if num_workers==0 else num_workers}
                cpu core(s) will transform the data.
                """,
            ))
        step.append(
            self.line("Click '+create'")
        )
        x_models = config["x_models"]
        y_models = config["y_models"]
        
        for xy, parts, color in zip(
            ["x","y"],
            [x_models, y_models],
            [self.x_color, self.y_color]):
            # going through model parts
            step.append(f"""
                <h4 id='step-model-{xy}' style='color:{color}'>{len(parts)}
                model parts for '{xy}'</h4>""")
            
            for col, model in parts.items():
                step.append(
                    self.line(f"""
                    <h6 id='step-model-{xy}-part-{col}' style='color:{color}'>
                    Select '{col}' as 'src'</h6>""")
                    )
                model_name = model['model_name']
                kwargs = model['kwargs']
                step.append(
                    self.line(f"""
                    Chooose <strong style="color:{color}">'{model_name}'
                    </strong>
                    """)
                )
                step.append(self.line("""
                Click 'Yes!', then set the kwargs to the following:"""))
                
                step.append(
                    self.line(list_group_kv(kwargs)))
                step.append(
                    self.line(f"Click 'Okay'")
                )
        step.append(self.ok("Congrats! now Click ✔️Next above"))
        return step
    
    def step_train(self) -> DOM:
        config = self.phase.config
        step = self.empty_step()
        step.append(self.title("Step 4: Training"))
        step.append("<h4>Final Step!</h4>")
        step.append("""
        🍹 You might experience some waiting while tai-chi-engine
        it's trying to download the pre-trained model""")
        step.append(self.ok(f"🚀 Click '+create' to start the training"))
        
        return step
        


from tai_chi_engine import TaiChiTrained
from tai_chi_engine.quantify import (
    QuantifyText,
    QuantifyCategory,
    QuantifyImage,
    QuantifyMultiCategory,
    QuantifyNum
)
import streamlit as st
from typing import Any, Dict, Callable
from pathlib import Path
import logging
import pandas as pd
from PIL import Image
import os


@st.cache(allow_output_mutation=True)
def load_trained(project: Path) -> TaiChiTrained:
    """
    Cache function to load medel
    """
    PROJECT = Path(project)
    logging.warning('loading... takes time')
    trained = TaiChiTrained(PROJECT)
    return trained


def get_enrich_map(trained: TaiChiTrained) -> Dict[str, Dict[str, Any]]:
    """
    Search the enrich configs by column names
    """
    if 'enrich' not in trained.phase.config:
        return dict()
    enrich_map = dict()
    for enrich in trained.phase['enrich']:
        enrich_map[enrich['dst']] = enrich
    return enrich_map


def to_build_st(QuantifyClass: type) -> Callable:
    """
    Assgin the QuantifyClass with build_st_field
    Use as decorator
    """
    def to_build_st_decorator(func: Callable) -> Callable:
        QuantifyClass.build_st_field = func
        return func
    return to_build_st_decorator


@to_build_st(QuantifyNum)
def st_num(quantify: QuantifyNum, name: str, enrich=None):
    """
    Build float number input slider
    """
    value = quantify.mean_
    min_value = quantify.mean_ - 3*quantify.std_
    max_value = quantify.mean_ + 3*quantify.std_

    return st.slider(
        label=name,
        min_value=min_value,
        max_value=max_value,
        value=value)


@to_build_st(QuantifyImage)
def st_image(quantify: QuantifyImage, name: str, enrich=None):
    """
    Build image upload field
    The result will be resized and converted
        according to the enrich config
    """
    if enrich is None:
        size = 224
        convert = "RGB"
    elif enrich['enrich'] != "EnrichImage":
        size = 224
        convert = "RGB"
    else:
        size = enrich['kwargs']['size']
        convert = enrich['kwargs']['convert']
    uploaded_file = st.file_uploader(
        label=name, type=["jpg", "png", "jpeg", "JPG", "PNG", "JPEG"])

    if uploaded_file is not None:
        image = Image.open(uploaded_file).resize((size, size))
        st.image(image, channels=convert)
        return image


@to_build_st(QuantifyText)
def st_text(quantify: QuantifyText, name: str, enrich=None):
    """
    Build text field based on max length
    if max length >64, use the text area
    if not, use the text input field
    """
    if quantify.max_length > 64:
        return st.text_area(label=name, )
    else:
        return st.text_input(label=name)


@to_build_st(QuantifyMultiCategory)
def st_multiselect(quantify: QuantifyMultiCategory, name: str, enrich=None):
    """
    Build multi-select based on the categories
    """
    val = st.multiselect(label=name, options=quantify.category.i2c)
    if quantify.actual_separator is not None:
        val = quantify.actual_separator.join(val)
    return val


@to_build_st(QuantifyCategory)
def st_select(quantify: QuantifyCategory, name: str, enrich=None):
    """
    Build select field for single categorical input
    """
    return st.selectbox(label=name, options=quantify.category.i2c)


def build_app(project: Path):
    """
    Build streamlit app
    with TaiChiTrained
    """
    # load a trained model
    trained = load_trained(Path(project))
    slug = trained.phase['task_slug']
    # build title
    st.markdown(f"""
    ## A Tai-Chi Engine Model:
    > Trained with [Tai-Chi Engine](https://github.com/tcengine/tai-chi)
    """)
    st.write(f'Task: {slug}')

    input_data = dict()
    enrich_map = get_enrich_map(trained)

    # build the input data
    for name, quantify in trained.qdict.items():
        # only build input upon x
        if quantify.is_x:
            if hasattr(quantify, 'build_st_field'):
                enrich = enrich_map.get(name)
                input_data[name] = quantify.build_st_field(
                    name, enrich)
            else:
                raise ValueError(
                    f"We don't have ways to build input field for column: {name}")

    # manage the predict output
    if st.button("Predict"):
        result = trained.predict(input_data)
        st.title("Result")
        if type(result) == pd.DataFrame:
            st.table(result.head(10))
        else:
            st.write(result)


if __name__ == "__main__":
    try:
        TAI_CHI_ENGINE_APP = os.environ['TAI_CHI_ENGINE_APP']
    except:
        print("Please set TAI_CHI_ENGINE_APP environment variable to the project directory")
        exit(1)

    build_app(TAI_CHI_ENGINE_APP)



__all__ = ["StartStreamLit", ]

from subprocess import Popen, PIPE
from tai_chi_engine.app.streamlit import run
import os
from io import StringIO
import sys


class StartStreamLit:
    def __init__(self, project: str, port: int = 8501):
        self.port = port
        self.process = None
        self.project = project
        self.run_script  = run.__file__
    
    def start(self):
        os.environ['TAI_CHI_ENGINE_APP'] = self.project
        self.process = Popen(
            ["streamlit", "run", self.run_script, "--server.port", str(self.port)],
            stdout=PIPE,
            stderr=PIPE,
        )

    def stop(self):
        self.process.terminate()
        self.process.wait()

    def follow(self):
        """
        Reads the output from the server process, and writes it to stdout.
        """
        for line in iter(self.process.stdout.readline, b''):
            sys.stdout.write(line.decode("utf-8"))


__all__ = ["Accuracy", "BiAccuracy", "F1Score", "MeanAbsoluteError"]

import torch
from typing import Dict

class Metric:
    def __init__(self):
        pass

    def __call__(self, y_, y):
        raise NotImplementedError(
            f"{self.__class__.__name__} is not defined properly")

class Accuracy(Metric):
    """
    Accuracy for multi-class classification
    """
    def __call__(self, y_, y):
        return (y_.argmax(-1) == y).float().mean()

def recall(y_pred, y_true, bm=0.5):
    is_right = (((y_pred > bm)*1).long() == y_true.long())
    is_right_in_targ = is_right[y_true == 1]
    if len(is_right_in_targ) == 0:
        is_right_in_targ = torch.zeros(1).to(y_pred.device)
    return is_right_in_targ.float().mean()


def precision(y_pred, y_true, bm=0.5):
    is_right = (((y_pred > bm)*1).long() == y_true.long())
    is_right_in_pred = is_right[(y_pred > bm)]
    if len(is_right_in_pred) == 0:
        is_right_in_pred = torch.zeros(1).to(y_pred.device)
    return is_right_in_pred.float().mean()

class BiAccuracy(Metric):
    def __call__(self, y_, y) -> Dict[str, float]:
        return dict(acc=((y_ > .5).float() == y).float().mean())

class F1Score(Metric):
    """
    F1 for binary classification
    """
    def __call__(self, y_, y) -> Dict[str, float]:
        rec = recall(y_, y)
        prec = precision(y_, y)
        denomenator = (rec + prec)
        if denomenator <= 0.:
            f1 = denomenator
        else:
            f1 = (2*rec*prec)/denomenator
        return dict(f1=f1, rec=rec, prec=prec)

class MeanAbsoluteError(Metric):
    """
    Mean Absolute Error
    """
    def __call__(self, y_, y):
        return dict(mae=torch.abs(y_ - y).mean())

    



__all__ = ['CategoryEncoder', 'MultiCategoryEncoder']

from tai_chi_tuna.front.typer import LIST
from .basic import EntryModel

from torch import nn

class CategoryEncoder(EntryModel):
    def __init__(self, num_embeddings, embedding_dim):
        super().__init__()
        self.model = nn.Embedding(
            num_embeddings,
            embedding_dim)
        
    def forward(self, idx):
        return self.model(idx)
    
    @classmethod
    def from_quantify(
        cls,
        quantify,
        embedding_dim: LIST(
            options=[4, 8, 16, 32, 64, 128, 256, 512], default=128) = 128):
        num_embeddings = len(quantify.category)
        obj = cls(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        obj.out_features = embedding_dim
        return obj
    
class MultiCategoryEncoder(EntryModel):
    def __init__(self, num_embeddings, embedding_dim):
        super().__init__()
        self.model = nn.Embedding(
            num_embeddings,
            embedding_dim)
        
    def forward(self, idx):
        return idx.float()@self.model.weight
    
    @classmethod
    def from_quantify(
        cls,
        quantify,
        embedding_dim: LIST(
            options=[4, 8, 16, 32, 64, 128, 256, 512], default=128) = 128):
        num_embeddings = len(quantify.category)
        obj = MultiCategoryEncoder(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        obj.out_features = embedding_dim
        return obj



from tai_chi_engine.quantify import (
    QuantifyCategory, QuantifyMultiCategory, QuantifyNum
)
__all__ = ['ExitModel', 'RegressionTop', 'CategoryTop',
           'MultiCategoryTop', 'ALL_EXIT', 'QUANTIFY_2_EXIT_MAP']

from .category import CategoryTop, MultiCategoryTop

from .basic import ExitModel, RegressionTop

ALL_EXIT = dict(
    CategoryTop=CategoryTop,
    MultiCategoryTop=MultiCategoryTop,
    RegressionTop=RegressionTop,
)

QUANTIFY_2_EXIT_MAP = dict({
    QuantifyCategory: [
        CategoryTop,
    ],
    QuantifyMultiCategory: [
        MultiCategoryTop,
    ],
    QuantifyNum: [
        RegressionTop,
    ],
})
